--
-- $Id: HrefDatabase.hs,v 1.9 2006/05/31 18:05:08 aamine Exp $
--
-- Copyright (c) 2003-2006 Minero Aoki
--
-- This program is free software.
-- You can distribute/modify this program under the terms of
-- the GNU LGPL, Lesser General Public License version 2.1.
-- For details of the GNU LPGL, see the file "COPYING".
--

module HrefDatabase
    (hrefDatabasePrefix, Entry (..), entContent, EntrySpec (..),
     specMatchBy, putEntry, putIndex,
     readModules, readSpecs, readModuleFunctions, readDocument) where

import PathUtils
import Utils
import URLEncoding
import Data.Char
import Data.List
import Data.Maybe
import System.Environment
#ifdef WIN32
import Win32
#endif

hrefDataDirEnv = "HREF_DATADIR"
hrefKeyListFile = "comptable"

hrefDatabasePrefix :: IO String
#if defined(POSIX)
hrefDatabasePrefix = getEnv hrefDataDirEnv
#elif defined(WIN32)
hrefDatabasePrefix = catch (getEnv hrefDataDirEnv)
                           (\ex -> do path <- exePath
                                      return (dirname path ++ "\\db"))

exePath = getModuleFileName =<< getModuleHandle Nothing
#endif

data Entry = Entry EntrySpec String

entContent (Entry s c) = c

entPath prefix (EntrySpec m f) = concatPath [prefix, encodePath m, encodePath f]

data EntrySpec = EntrySpec String String

instance Show EntrySpec where
  show (EntrySpec m f) = m ++ "." ++ escapeFunction f

escapeFunction s
    | isJust $ find (not . isAlphaNum) s  = "(" ++ s ++ ")"
    | otherwise = s

serializeSpec (EntrySpec m f) = m ++ "," ++ f

parseSpec s = let (m,f) = break (== ',') s
              in EntrySpec m (tail f)

specMatchBy :: String -> String -> (String -> String -> Bool) -> EntrySpec -> Bool
specMatchBy ""   ""   f _                    = True
specMatchBy ""   fpat f (EntrySpec _ func)   = f fpat func
specMatchBy mpat ""   f (EntrySpec mod _)    = f mpat mod
specMatchBy mpat fpat f (EntrySpec mod func) = f mpat mod && f fpat func

putEntry :: FilePath -> Entry -> IO ()
putEntry prefix (Entry s c) =
    do makePath (dirname $ entPath prefix s)
       writeFile (entPath prefix s) $ c

putIndex :: FilePath -> [Entry] -> IO ()
putIndex prefix ents =
    do makePath prefix
       writeFile (comptablePath prefix) $ unlines $ map entIndex ents
  where
    comptablePath prefix = joinPath prefix hrefKeyListFile

    entIndex (Entry s c) = serializeSpec s

readModules prefix =
    return . map decodePath . filter (any isUpper) =<< dirEntries prefix

readSpecs :: FilePath -> IO [EntrySpec]
readSpecs prefix =
    return . map parseSpec . lines
    =<< readFile (joinPath prefix hrefKeyListFile)

readModuleFunctions :: FilePath -> String -> IO [String]
readModuleFunctions prefix mod =
    return . map decodePath =<< dirEntries (joinPath prefix $ encodePath mod)

readDocument :: FilePath -> EntrySpec -> IO String
readDocument prefix spec = readFile (entPath prefix spec)

encodePath = urlencode
decodePath = urldecode
